#By defaualt, smallab will run your experiments in joblib using threads
#This can be a bit slower than using multiprocessing
import logging
import time
import typing

import random

from smallab.experiment import Experiment
from smallab.runner import ExperimentRunner
from smallab.runner_implementations.multiprocessing_runner import MultiprocessingRunner
from smallab.specification_generator import SpecificationGenerator
from examples.example_utils import delete_experiments_folder


#Similar experiment as before but now it will sleep for a bit, taking longer to complete in serial
class SimpleExperiment(Experiment):
    def main(self, specification: typing.Dict) -> typing.Dict:
        logging.getLogger(self.get_logger_name()).info("Doing work!")
        random.seed(specification["seed"])
        for i in range(specification["num_calls"]):
            r = random.random()
        time.sleep(10 * r)
        return {"number": r}


#Same specification as before
generation_specification = {"seed": [1, 2, 3, 4, 5, 6, 7, 8], "num_calls": [1, 2, 3]}
specifications = SpecificationGenerator().generate(generation_specification)

name = "multiprocessing_runner"
#This time we will run them all in parallel
runner = ExperimentRunner()
runner.run(name, specifications, SimpleExperiment(),specification_runner=MultiprocessingRunner())

#Cleanup example
delete_experiments_folder(name)



#This is an example of a very simple experiment in smallab
# What you need to import from smallab
import logging

import random
import typing

from examples.example_utils import delete_experiments_folder
from smallab.experiment import Experiment
from smallab.runner import ExperimentRunner


# Write a simple experiment
class SimpleExperiment(Experiment):
    # Need to implement this method, will be passed the specification
    # Return a dictionary of results
    def main(self, specification: typing.Dict) -> typing.Dict:
        random.seed(specification["seed"])
        for i in range(specification["num_calls"]):  # Advance the random number generator some amount
            random.random()
        #return the random number. This, along with the specification that generated it will be saved
        return {"number": random.random()}

#The name describes what experiment your doing
name = "simple_experiment1"
#The specifications are a list of dictionaries that will get passed to your experiment instance
specifications = [{"seed":1,"num_calls":1},{"seed":2,"num_calls":2}]
#This object is the code to run your experiment, an instance of the Experiment class
experiment_instance = SimpleExperiment()

#The experiment runner handles cloning your experiment instance and giving it a specification to run with
runner = ExperimentRunner()
runner.run(name,specifications,experiment_instance)

#If I run it again, nothing will happen because smallab knows those experiments succesfully completed
runner.run(name,specifications,experiment_instance)
delete_experiments_folder(name)



import email
import json
import smtplib
import typing

from smallab.callbacks import CallbackManager
from smallab.utilities.hooks import format_exception


class EmailCallback(CallbackManager):
    def __init__(self, email_adress, port):
        self.email_address = email_adress
        self.smtp_port = port

    def __send(self, message):
        message["Subject"] = "Smallab report"
        message["From"] = self.email_address
        message["To"] = self.email_address
        s = smtplib.SMTP('localhost', self.smtp_port)
        s.send_message(message)
        s.quit()

    def on_specification_failure(self, exception, specification):
        message = email.message.EmailMessage()
        content = "!!! Failure !!! \n\n Specification: %s \n\n Exception: %s" % (
            str(json.dumps(specification, sort_keys=True, indent=1)), str(format_exception(exception)))
        message.set_content(content)
        self.__send(message)

    def on_specification_complete(self, specification: typing.Dict, result: typing.Dict) -> typing.NoReturn:
        message = email.message.EmailMessage()
        content = "!!! Sucesss !!! \n\n Specification: %s" % (str(json.dumps(specification, sort_keys=True, indent=1)))
        message.set_content(content)
        self.__send(message)

    def on_batch_failure(self, errors: typing.List[Exception],
                         specifications: typing.List[typing.Dict]) -> typing.NoReturn:
        message = email.message.EmailMessage()

        content = "!!! Failure !!! \n\n "
        for exception, specification in zip(errors, specifications):
            content += "Specification: %s \n\n Exception: %s \n\n" % (
                str(json.dumps(specification, sort_keys=True, indent=1)), str(format_exception(exception)))
        message.set_content(content)

        self.__send(message)

    def on_batch_complete(self, specifications: typing.List[typing.Dict]) -> typing.NoReturn:
        message = email.message.EmailMessage()
        content = "!!! Sucesss !!! \n\n"
        for specification in specifications:
            content += "Specification: %s \n\n" % (str(json.dumps(specification, sort_keys=True, indent=1)))
        message.set_content(content)
        self.__send(message)


class EmailCallbackBatchOnly(EmailCallback):
    def on_specification_complete(self, specification: typing.Dict, result: typing.Dict) -> typing.NoReturn:
        return None

    def on_batch_failure(self, errors: typing.List[Exception],
                         specifications: typing.List[typing.Dict]) -> typing.NoReturn:
        return None



import json

import hashlib
import humanhash


def specification_hash(specification):
    return humanhash.humanize(hashlib.md5(json.dumps(specification, sort_keys=True).encode("utf-8")).hexdigest())



import typing

import abc

from smallab.smallab_types import Specification


class AbstractRunner(abc.ABC):
    """
    The base class for running a batch of specifications.

    The subclass must implement run and call finish after it is run.
    """

    def finish(self, completed_specifications: typing.List[Specification],
               failed_specifications: typing.List[Specification], exceptions: typing.List[Exception]):
        """
        Call in .run to save the results of the batch of specification

        It is assumed that the ith failed_specification and the ith exception are from the same experiment
        :param completed_specifications: Specifications that completed succesfully (returned None)
        :param failed_specifications: Specifications that failed  (retuned an Exception)
        :param exceptions: The exceptions corresponding to the failed exceptions
        :return:
        """
        self.completed_specifications = completed_specifications
        self.failed_specifications = failed_specifications
        self.exceptions = exceptions

    @abc.abstractmethod
    def run(self, specifications_to_run: typing.List[Specification],
            run_and_save_fn: typing.Callable[[Specification], typing.Union[None, Exception]]):
        """
        The method to be implemented in concrete runner clases
        :param specifications_to_run: The list of specifications to run, all specifications should be run
        :param run_and_save_fn: The function to run and save the specifications, Each specification should have this function called on it. Returns None for a successful completion and Exception on failure.
        """
        pass

    def get_completed(self):
        return self.completed_specifications

    def get_failed_specifications(self):
        return self.failed_specifications

    def get_exceptions(self):
        return self.exceptions



from queue import Full

from smallab.dashboard.dashboard import eventQueue
from smallab.dashboard.dashboard_events import LogEvent


class LogToEventQueue:
    """A file-like object that writes to the event queue.
     This is used so that logs will get added to the event queue to be displayed by the dashboard"""

    def write(self, t):
        put_in_event_queue(LogEvent(t))

    def flush(self):
        pass

def put_in_event_queue(o):
    '''
    This is just to put in the event queue without waiting or failing if it's full
    :param o:
    :return:
    '''
    try:
        eventQueue.put_nowait(o)
    except Full:
        pass



# This demonstrates using the specification generator to save writing each specification out
import typing

import random

from examples.example_utils import delete_experiments_folder
from smallab.experiment import Experiment
from smallab.runner import ExperimentRunner
from smallab.specification_generator import SpecificationGenerator


# Same experiment as before
class SimpleExperiment(Experiment):
    def main(self, specification: typing.Dict) -> typing.Dict:
        random.seed(specification["seed"])
        for i in range(specification["num_calls"]):
            random.random()
        return {"number": random.random()}


# In the generation specification keys that have lists as their values will be cross producted with other list valued keys to create many specifications
# in this instance there will be 8 * 3 = 24 specifications
generation_specification = {"seed": [1, 2, 3, 4, 5, 6, 7, 8], "num_calls": [1, 2, 3]}

# Call the generate method. Will create the cross product.
specifications = SpecificationGenerator().generate(generation_specification)
print(specifications)

name = "specification_generation_experiment"
runner = ExperimentRunner()
runner.run(name, specifications, SimpleExperiment())

delete_experiments_folder(name)


import curses
import logging
import math
import multiprocessing as mp
import time

import numpy as np

from smallab.dashboard.dashboard_events import (BeginEvent, CompleteEvent, ProgressEvent, LogEvent,
                                                StartExperimentEvent,
                                                RegisterEvent, FailedEvent)

#This is a globally accessible queue which stores the events for the dashboard
eventQueue = mp.Queue(maxsize=200)


class TimeEstimator():
    """
    This is used to record how long it takes for one iteration of an experiment and how long it takes to complete an experiment
    It uses this and the number of remaining experiments and iterations to compute the median and lower and upper quartile
    time to complete the entire batch
    """
    def __init__(self):
        self.last_update_time = dict()
        self.start_time = dict()
        self.time_per_iterations = []
        self.time_per_completion = []
        self.last_progress = dict()

    def record_start(self, specification):
        self.start_time[specification] = time.time()
        self.last_update_time[specification] = time.time()

    def record_iteration(self, specification,progress):
        if specification not in self.last_progress:
            #TODO this will likely be wrong when resuming, it assumes the progress started at 0
            progress_amount = progress
            self.last_progress[specification] = progress
        else:
            progress_amount = progress - self.last_progress[specification]
            self.last_progress[specification] = progress
        current_time = time.time()
        time_diff = current_time - self.last_update_time[specification]
        self.last_update_time[specification] = current_time
        self.time_per_iterations.append(time_diff / progress_amount)

    def record_completion(self, specification):
        current_time = time.time()
        time_diff = current_time - self.start_time[specification]
        self.time_per_completion.append(time_diff)
        self.last_update_time[specification] = current_time

    def compute_time_stats(self, specification_progress, active,registered):
        # Calculate this differently it should be remaining_progress + active_not_in_progress
        # Dictionary is empty, use number active and completion time to estimate
        # The math here is a little wonky since idk if median and quartiles work with iterated expectation

        expected_seconds_to_complete_all = 0
        lower_quartile_complete_all = 0
        higher_quartile_complete_all = 0
        if self.time_per_completion:
            times_to_complete = np.array(self.time_per_completion)
            #this finds the number experiments with no progress (so they are not double counted when counting iterations remaining)
            remaining_amount_with_no_progress = len(registered) + len(list(filter(lambda x: not x in specification_progress,active)))

            average_seconds_to_complete = np.median(times_to_complete)
            expected_seconds_to_complete_all = average_seconds_to_complete * remaining_amount_with_no_progress
            lower_quartile_complete_all = np.quantile(average_seconds_to_complete,
                                                      .25) * remaining_amount_with_no_progress
            higher_quartile_complete_all = np.quantile(average_seconds_to_complete,
                                                       .75) * remaining_amount_with_no_progress

        if self.time_per_iterations:
            times_per_iteration = np.array(self.time_per_iterations)
            remaining_iterations = sum(
                [max_num - current_progress for current_progress, max_num in specification_progress.values()])
            if not self.time_per_completion:
                remaining_amount_with_no_progress = len(registered) + len(
                    list(filter(lambda x: not x in specification_progress, active)))
                average_max = np.mean([max_num for current_progress,max_num in specification_progress.values()])
                remaining_iterations += average_max * remaining_amount_with_no_progress

            expected_seconds_to_complete_all += np.median(times_per_iteration) * remaining_iterations
            lower_quartile_complete_all += np.quantile(times_per_iteration,
                                                      .25) * remaining_iterations
            higher_quartile_complete_all += np.quantile(times_per_iteration,
                                                       .75) * remaining_iterations

        #Divide by active number since that's the amount of processes running
        return expected_seconds_to_complete_all / len(active), lower_quartile_complete_all / len(active), higher_quartile_complete_all / len(active)


intervals = (
    ('weeks', 604800),  # 60 * 60 * 24 * 7
    ('days', 86400),  # 60 * 60 * 24
    ('hours', 3600),  # 60 * 60
    ('minutes', 60),
    ('seconds', 1),
)


def display_time(seconds, granularity=2):
    result = []

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip('s')
            result.append("{} {}".format(value, name))
    return ', '.join(result[:granularity])


def draw_header_widget(row, stdscr, experiment_name, width, complete, active, registered, specification_progress,
                       timeestimator,failed):
    stdscr.addstr(0, 0, "Smallab Running: {name}".format(name=experiment_name))
    row += 1
    stdscr.addstr(row, 0, "-" * width)
    row += 1
    completed_failed_string = "Completed {num_complete} / {num_total}".format(num_complete=len(complete),
                                                                          num_total=len(complete) + len(active) + len(
                                                                              registered))
    if failed:
        completed_failed_string += " - Failed: {num_failed}".format(num_failed=len(failed))
    stdscr.addstr(row, 0,completed_failed_string)
    t = timeestimator.compute_time_stats(specification_progress, active,registered)
    if t is not None:
        expected, lower, higher = t
        completion_stats_string = "Completion - Expected: {expectedstr} - Upper: {upperstr} - Lower: {lowerstr}".format(
            expectedstr=display_time(expected), upperstr=display_time(higher), lowerstr=display_time(lower))
        stdscr.addstr(row, width - len(completion_stats_string), completion_stats_string)
    row += 1
    stdscr.addstr(row, 0, "-" * width)
    row += 1
    return row


def draw_specifications_widget(row, stdscr, active, registered, width, specification_progress, height,failed):
    start_row = row
    # Decide to draw in single or double column
    second_column_begins = math.floor(width / 2)
    max_height = math.floor(height / 2)
    use_double_column_layout = width >= 40 and len(active) + len(registered) > max_height - row
    on_second_column = False
    for i, active_specification in enumerate(active + registered+ failed):
        if use_double_column_layout and on_second_column:
            stdscr.addstr(row, second_column_begins,active_specification)
        else:
            stdscr.addstr(row, 0, active_specification)
        if use_double_column_layout:
            bar_width = math.floor(width / 8)
        else:
            bar_width = math.floor(width / 4)

        if active_specification in specification_progress:
            progress, max_amount = specification_progress[active_specification]
            status_string = "{progress}/{max_amount} : ".format(progress=progress, max_amount=max_amount)
            amount_complete = progress / max_amount
            bars_complete = math.floor(amount_complete * bar_width)
            bars_not_complete = bar_width - bars_complete
            status_string += "=" * bars_complete
            status_string += " " * bars_not_complete

        elif active_specification in active:
            status_string = "Running..."
        elif active_specification in failed:
            status_string = "Failed!"
        else:
            status_string = "Waiting..."
        if use_double_column_layout and not on_second_column:
            status_string += "||"
            stdscr.addstr(row, second_column_begins - len(status_string), status_string)
        else:
            stdscr.addstr(row, width - len(status_string), status_string)

        row += 1
        if row >= max_height:
            if use_double_column_layout and not on_second_column:
                on_second_column = True
                row = start_row
            else:
                break
    assert row <= max_height
    return max_height


def draw_log_widget(row, stdscr, width, height, log_spool):
    stdscr.addstr(row, 0, "=" * width)
    row += 1
    stdscr.addstr(row, 0, "Logs")
    row += 1
    stdscr.addstr(row, 0, "=" * width)
    row += 1
    remaining_rows = height - row
    for log in log_spool[-remaining_rows + 1:]:
        message = log[:width]
        if not message.isspace():
            stdscr.addstr(row, 0, message)
            row += 1
    return row


def run(stdscr):
    timeestimator = TimeEstimator()
    i = 0
    log_spool = []
    active = []
    complete = []
    experiment_name = ""
    specification_progress = dict()
    registered = []
    failed = []
    while True:
        # Drain Queue:
        try:
            while not eventQueue.empty():
                event = eventQueue.get()
                if isinstance(event, BeginEvent):
                    registered.remove(event.specification_id)
                    active.append(event.specification_id)
                    timeestimator.record_start(event.specification_id)
                elif isinstance(event, CompleteEvent):
                    active.remove(event.specification_id)
                    complete.append(event.specification_id)
                    timeestimator.record_completion(event.specification_id)
                elif isinstance(event, ProgressEvent):
                    specification_progress[event.specification_id] = (event.progress, event.max)
                    timeestimator.record_iteration(event.specification_id,event.progress)
                elif isinstance(event, LogEvent):
                    if not event.message.isspace():
                        log_spool.append(event.message)
                elif isinstance(event, StartExperimentEvent):
                    experiment_name = event.name
                elif isinstance(event, RegisterEvent):
                    registered.append(event.specification_id)
                elif isinstance(event,FailedEvent):
                    active.remove(event.specification_id)
                    failed.append(event.specification_id)
                else:
                    print("Dashboard action not understood")

            # Draw Screen
            stdscr.clear()
            height, width = stdscr.getmaxyx()
            row = 0
            row = draw_header_widget(row, stdscr, experiment_name, width, complete, active, registered,
                                     specification_progress, timeestimator,failed)
            row = draw_specifications_widget(row, stdscr, active, registered, width, specification_progress, height,failed)
            row = draw_log_widget(row, stdscr, width, height, log_spool)
            stdscr.refresh()
            time.sleep(0.1)
        except Exception as e:
            logging.getLogger("smallab.dashboard").error("Dashboard Error {}".format(e),exc_info=True)



def start_dashboard():
    curses.wrapper(run)




import logging

import random
import typing

from examples.example_utils import delete_experiments_folder
from smallab.experiment import Experiment
from smallab.runner import ExperimentRunner


#This is another simple experiment, this time with logging involved!
class SimpleExperiment(Experiment):
    def main(self, specification: typing.Dict) -> typing.Dict:
        #Get logger name is set per experiment, all logging here will go to a seperate file with the experiment name and go to a main log as well
        logging.getLogger(self.get_logger_name()).info("Doing work!")
        random.seed(specification["seed"])
        for i in range(specification["num_calls"]):  # Advance the random number generator some amount
            random.random()
        return {"number": random.random()}


name = "simple_experiment2"
runner = ExperimentRunner()
runner.run(name,[{"seed":1,"num_calls":1},{"seed":2,"num_calls":2}],SimpleExperiment())

delete_experiments_folder(name)



import traceback
import typing


def format_exception(e: Exception) -> typing.AnyStr:
    # https://stackoverflow.com/questions/9555133/e-printstacktrace-equivalent-in-python
    return "".join(traceback.format_exception(None, e, e.__traceback__))



import datetime
import json
import logging
import multiprocessing as mp
import typing

import dill
import os
from copy import deepcopy

from smallab.callbacks import CallbackManager
from smallab.checkpointed_experiment_handler import CheckpointedExperimentHandler
from smallab.dashboard.dashboard import start_dashboard
from smallab.dashboard.dashboard_events import BeginEvent, CompleteEvent, StartExperimentEvent, RegisterEvent, \
    FailedEvent
from smallab.dashboard.utils import LogToEventQueue, put_in_event_queue
from smallab.experiment import CheckpointedExperiment, BaseExperiment
from smallab.file_locations import (get_save_file_directory, get_json_file_location, get_pkl_file_location,
                                    get_specification_file_location, get_save_directory, get_experiment_save_directory,
                                    get_log_file)
from smallab.runner_implementations.abstract_runner import AbstractRunner
from smallab.runner_implementations.joblib_runner import JoblibRunner
from smallab.smallab_types import Specification
from smallab.specification_hashing import specification_hash
from smallab.utilities.logging_callback import LoggingCallback


class ExperimentRunner(object):
    """
    The class which runs the batch of specifications.
    """

    def __init__(self):
        self.callbacks = [LoggingCallback()]
        self.logging_callback_attached = False

    def attach_callbacks(self, callbacks: typing.List[CallbackManager]):
        """
        Attach callback handlers to this runner object, will call them in order they are presented
        :return:
        """
        self.logging_callback_attached = False
        for callback in callbacks:
            if isinstance(callback, LoggingCallback):
                self.logging_callback_attached = True
        self.callbacks = callbacks

    def set_experiment_folder(self, folder: typing.AnyStr) -> typing.NoReturn:
        """
        Sets the folder where experiments will be stored
        defaults to "experiment_runs/"

        :param folder: A String folder path for the experiments
        :return: Nothing
        """
        self.experiment_folder = folder

    def _write_to_completed_json(self, name: typing.AnyStr, completed_specifications: typing.List[Specification],
                                 failed_specifications: typing.List[Specification]):
        """
        Writes out a file which shows the completed and failed specifications
        :param name: The name of the current batch
        :param completed_specifications:  Specifications which completed successfully
        :param failed_specifications:  Specifications which failed
        """
        with open(os.path.join(get_save_directory(name), "completed.json"), 'w') as f:
            json.dump(completed_specifications, f)
        with open(os.path.join(get_save_directory(name), "failed.json"), 'w') as f:
            json.dump(failed_specifications, f)

    def _find_uncompleted_specifications(self, name, specifications):
        already_completed_specifications = []
        for root, _, files in os.walk(get_experiment_save_directory(name)):
            for fname in files:
                if ".pkl" in fname:
                    with open(os.path.join(root, fname), "rb") as f:
                        completed = dill.load(f)
                    already_completed_specifications.append(completed["specification"])
                if ".json" in fname and fname != 'specification.json':
                    with open(os.path.join(root, fname), "r") as f:
                        completed = json.load(f)
                    already_completed_specifications.append(completed["specification"])

        need_to_run_specifications = []
        for specification in specifications:
            if specification in already_completed_specifications:
                logging.getLogger("smallab.runner").info("Skipping: " + str(specification))
            else:
                need_to_run_specifications.append(specification)
        return need_to_run_specifications

    def run(self, name: typing.AnyStr, specifications: typing.List[Specification], experiment: BaseExperiment,
            continue_from_last_run=True, propagate_exceptions=False,
            force_pickle=False, specification_runner: AbstractRunner = None, use_dashboard=True) -> typing.NoReturn:
        """
        The method called to run an experiment
        :param propagate_exceptions: If True, exceptions won't be caught and logged as failed experiments but will cause the program to crash (like normal), useful for debugging exeperiments
        :param name: The name of this experiment batch
        :param specifications: The list of specifications to run. Should be a list of dictionaries. Each dictionary is passed to the experiment run method
        :param experiment: The experiment object to run
        :param continue_from_last_run: If true, will not redo already completed experiments. Defaults to true
        :param show_progress: Whether or not to show a progress bar for experiment completion
        :param force_pickle: If true, don't attempt to json serialze results and default to pickling
        :param specification_runner: An instance of ```AbstractRunner``` that will be used to run the specification
        :param use_dashboard: If true, use the terminal monitoring dashboard. If false, just stream logs to stdout.
        :return: No return
        """
        if specification_runner is None:
            specification_runner = JoblibRunner(None)
        dashboard_process = None
        try:
            put_in_event_queue(StartExperimentEvent(name))
            # Set up root smallab logger
            folder_loc = os.path.join("experiment_runs", name, "logs", str(datetime.datetime.now()))
            file_loc = os.path.join(folder_loc, "main.log")
            if not os.path.exists(folder_loc):
                os.makedirs(folder_loc)
            logger = logging.getLogger("smallab")
            logger.setLevel(logging.DEBUG)
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            fh = logging.FileHandler(file_loc)
            fh.setFormatter(formatter)
            logger.addHandler(fh)
            if not use_dashboard:
                sh = logging.StreamHandler()
                sh.setFormatter(formatter)
                logger.addHandler(sh)
            else:
                fq = LogToEventQueue()
                sh = logging.StreamHandler(fq)
                sh.setFormatter(formatter)
                logger.addHandler(sh)
                dashboard_process = mp.Process(target=start_dashboard)
                dashboard_process.start()
            experiment.set_logging_folder(folder_loc)

            self.force_pickle = force_pickle
            if not os.path.exists(get_save_directory(name)):
                os.makedirs(get_save_directory(name))

            if continue_from_last_run:
                need_to_run_specifications = self._find_uncompleted_specifications(name, specifications)
            else:
                need_to_run_specifications = specifications
            for callback in self.callbacks:
                callback.set_experiment_name(name)

            for specification in need_to_run_specifications:
                put_in_event_queue(RegisterEvent(specification_hash(specification)))
            specification_runner.run(need_to_run_specifications,
                                     lambda specification: self.__run_and_save(name, experiment, specification,
                                                                               propagate_exceptions))
            self._write_to_completed_json(name, specification_runner.get_completed(),
                                          specification_runner.get_failed_specifications())

            # Call batch complete functions
            if specification_runner.get_exceptions() != []:
                for callback in self.callbacks:
                    callback.on_batch_failure(specification_runner.get_exceptions(),
                                              specification_runner.get_failed_specifications())

            if specification_runner.get_completed() != []:
                for callback in self.callbacks:
                    callback.on_batch_complete(specification_runner.get_completed())
        finally:
            if dashboard_process is not None:
                dashboard_process.terminate()

    def __run_and_save(self, name, experiment, specification, propagate_exceptions):
        experiment = deepcopy(experiment)
        specification_id = specification_hash(specification)
        logger_name = "smallab.{specification_id}".format(specification_id=specification_id)
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.DEBUG)
        file_handler = logging.FileHandler(get_log_file(experiment, specification_id))
        formatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s]  %(message)s")
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        experiment.set_logger_name(logger_name)
        put_in_event_queue(BeginEvent(specification_id))

        def _interior_fn():
            if isinstance(experiment, CheckpointedExperiment):
                result = CheckpointedExperimentHandler().run(experiment, name, specification)
            else:
                result = experiment.main(specification)
            self._save_run(name, experiment, specification, result)
            for callback in self.callbacks:
                callback.on_specification_complete(specification, result)
            return None

        if not propagate_exceptions:
            try:
                _interior_fn()

                put_in_event_queue(CompleteEvent(specification_id))
            except Exception as e:
                logger.error("Specification Failure", exc_info=True)
                put_in_event_queue(FailedEvent(specification_id))
                for callback in self.callbacks:
                    callback.on_specification_failure(e, specification)
                return e
        else:
            _interior_fn()
            put_in_event_queue(CompleteEvent(specification_id))
            return None

    def _save_run(self, name, experiment, specification, result):
        os.makedirs(get_save_file_directory(name, specification))
        output_dictionary = {"specification": specification, "result": result}
        json_serialize_was_successful = False
        # Try json serialization
        if not self.force_pickle:
            json_filename = get_json_file_location(name, specification)
            try:
                with open(json_filename, "w") as f:
                    json.dump(output_dictionary, f)
                json_serialize_was_successful = True
            except Exception:
                logging.getLogger(experiment.get_logger_name()).warning("Json serialization failed with exception",
                                                                        exc_info=True)
                os.remove(json_filename)
        # Try pickle serialization
        if self.force_pickle or not json_serialize_was_successful:
            pickle_file_location = get_pkl_file_location(name, specification)
            specification_file_location = get_specification_file_location(name, specification)
            try:
                with open(pickle_file_location, "wb") as f:
                    dill.dump(output_dictionary, f)
                with open(specification_file_location, "w") as f:
                    json.dump(specification, f)
            except Exception:
                logging.getLogger(experiment.get_logger_name()).critical("Experiment results serialization failed!!!",
                                                                         exc_info=True)
                try:
                    os.remove(pickle_file_location)
                except FileNotFoundError:
                    pass
                try:
                    os.remove(specification_file_location)
                except FileNotFoundError:
                    pass



import logging
import typing

from joblib import Parallel, delayed
from tqdm import tqdm

from smallab.runner_implementations.abstract_runner import AbstractRunner
from smallab.smallab_types import Specification
from smallab.utilities.tqdm_to_logger import TqdmToLogger


class JoblibRunner(AbstractRunner):
    """
    Runs the Specifications in parallel on threads using joblib
    """

    def __init__(self, num_parallel):
        self.num_parallel = num_parallel

    def run(self, specifications_to_run: typing.List[Specification],
            run_and_save_fn: typing.Callable[[Specification], typing.Union[None, Exception]]):
        completed_specifications = []
        failed_specifications = []
        exceptions = []
        with tqdm(total=len(specifications_to_run), file=TqdmToLogger(logging.getLogger("smallab.joblib_runner")),
                  desc="Running Specifications") as pbar:
            def parallel_f(specification):
                run_and_save_fn(specification)
                pbar.update(1)

            exceptions_thrown = Parallel(n_jobs=self.num_parallel, prefer="threads")(
                delayed(lambda specification: parallel_f(specification))(specification) for
                specification in specifications_to_run)

        # Look through output to create batch failures and sucesses
        for specification, exception_thrown in zip(specifications_to_run, exceptions_thrown):
            if exception_thrown is None:
                completed_specifications.append(specification)
            else:
                exceptions.append(exception_thrown)
                failed_specifications.append(specification)
        self.finish(completed_specifications, failed_specifications, exceptions)



import logging
import typing

from tqdm import tqdm

from smallab.runner_implementations.abstract_runner import AbstractRunner
from smallab.smallab_types import Specification
from smallab.utilities.tqdm_to_logger import TqdmToLogger


class MainRunner(AbstractRunner):
    """
    The simplest runner which runs each specification in serial on the main process and thread.
    """

    def __init__(self, show_progress=True):
        super().__init__()
        self.show_progress = show_progress

    def run(self, specifications_to_run: typing.List[Specification], run_and_save_fn):
        completed_specifications = []
        failed_specifications = []
        exceptions = []
        for specification in tqdm(specifications_to_run, file=TqdmToLogger(logging.getLogger("smallab.main_process_runner")),
                                  desc="Experiments", disable=not self.show_progress):
            exception_thrown = run_and_save_fn(specification)

            if exception_thrown is None:
                completed_specifications.append(specification)
            else:
                failed_specifications.append(specification)
                exceptions.append(exception_thrown)
        self.finish(completed_specifications, failed_specifications, exceptions)



import datetime
import logging
import typing

import dill

from smallab.runner_implementations.abstract_runner import AbstractRunner
from smallab.smallab_types import Specification
from smallab.specification_hashing import specification_hash


def run_dill_encoded(payload):
    fun, args = dill.loads(payload)
    return fun(*args)


def apply_async(pool, fun, args):
    payload = dill.dumps((fun, args))
    return pool.apply_async(run_dill_encoded, (payload,))


def parallel_f(specification, run_and_save_fn, idx, max_num, counter, seconds_completion):
    alpha = .8
    specification_id = specification_hash(specification)
    logger = logging.getLogger("smallab.{id}.multiprocessing_runner".format(id=specification_id))
    time_str = "%d-%b-%Y (%H:%M:%S.%f)"
    beginning = datetime.datetime.now()
    beginning_string = beginning.strftime(time_str)
    logger.info("Begin job {idx} / {max_num} @ {beginning_string}".format(idx=idx, max_num=max_num,
                                                                          beginning_string=beginning_string))
    result = run_and_save_fn(specification)

    if not isinstance(result, Exception):
        end = datetime.datetime.now()
        end_str = end.strftime(time_str)
        time_to_complete = (end - beginning).total_seconds()
        if seconds_completion.get() == 0.0:
            seconds_completion.set(time_to_complete)
        else:
            seconds_completion.set(alpha * time_to_complete + (1 - alpha) * seconds_completion.get())
        completed_by = datetime.datetime.now() + datetime.timedelta(0, (
                max_num - counter.value) * seconds_completion.get())
        counter.value += 1
        logger.info(
            "Completed {counter}/{max_num} total @ {end_str} Taking {time_to_complete}. Estimated per job {seconds_completion}s. Estimate to complete @ {completed_by}.".format(
                counter=counter.value, max_num=max_num, end_str=end_str, time_to_complete=time_to_complete,
                seconds_completion=seconds_completion.get(), completed_by=completed_by))
    return result


class MultiprocessingRunner(AbstractRunner):
    """
    A runner which uses a multiprocessing pool to manage specification running
    """

    def __init__(self, mp_override=None, num_parallel=None):
        """
        :param mp_override: provide multiprocessing library that should be used. This is done because pytorch has a funky multiprocessing library that could be pased in here
        """
        if mp_override is not None:
            self.mp = mp_override
        else:
            import multiprocessing
            self.mp = multiprocessing
        self.num_parallel = None

    def run(self, specifications_to_run: typing.List[Specification],
            run_and_save_fn: typing.Callable[[Specification], typing.Union[None, Exception]]):
        pool = self.mp.Pool(self.num_parallel)
        manager = self.mp.Manager()
        counter = manager.Value('i', 0)
        seconds_completion = manager.Value("d", 0)

        def interior_fn(specification):
            return parallel_f(specification, run_and_save_fn, idx + 1, max_num, counter, seconds_completion)

        # listener.start()
        jobs = []
        max_num = len(specifications_to_run)
        for idx, specification in enumerate(specifications_to_run):
            jobs.append(apply_async(pool, interior_fn, (specification,)))
        results = []
        for job in jobs:
            results.append(job.get())
        completed_specifications = []
        exceptions = []
        failed_specifications = []
        for specification, exception_thrown in zip(specifications_to_run, results):
            if exception_thrown is None:
                completed_specifications.append(specification)
            else:
                exceptions.append(exception_thrown)
                failed_specifications.append(specification)
        self.finish(completed_specifications, failed_specifications, exceptions)



class LogEvent():
    def __init__(self, message):
        self.message = message


class BeginEvent():
    def __init__(self, specification_id):
        self.specification_id = specification_id


class ProgressEvent():
    def __init__(self, specification_id, progress, max):
        self.specification_id = specification_id
        self.progress = progress
        self.max = max


class CompleteEvent():
    def __init__(self, specification_id):
        self.specification_id = specification_id


class StartExperimentEvent():
    def __init__(self, name):
        self.name = name


class RegisterEvent():
    def __init__(self, specification_id):
        self.specification_id = specification_id
class FailedEvent():
    def __init__(self,specification_id):
        self.specification_id = specification_id



import logging
import logging.config
import pprint
import traceback
import typing

from smallab.callbacks import CallbackManager


class LoggingCallback(CallbackManager):
    @staticmethod
    def pretty_printer():
        return pprint.PrettyPrinter()

    def on_specification_complete(self, specification: typing.Dict, result: typing.Dict) -> typing.NoReturn:
        pp = LoggingCallback.pretty_printer()
        specification_string = pp.pformat(specification)
        result_string = pp.pformat(result)
        logging.info(
            "\nSpecification Complete\nSpecification:\n{}\nResult:\n{}".format(specification_string,
                                                                               result_string))

    def on_specification_failure(self, exception: Exception, specification: typing.Dict) -> typing.NoReturn:
        pp = LoggingCallback.pretty_printer()
        traceback_string = traceback.format_exc()
        specification_string = pp.pformat(specification)
        logging.error("\nSpecification Failure\nSpecification:\n{}\nException:\n{}".format(specification_string,
                                                                                           traceback_string))



import logging

import io


class TqdmToLogger(io.StringIO):
    """
        Output stream for TQDM which will output to logger module instead of
        the StdOut.
    """
    logger = None
    level = None
    buf = ''

    def __init__(self, logger, level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = logger
        self.level = level or logging.INFO

    def write(self, buf):
        self.buf = buf.strip('\r\n\t ')

    def flush(self):
        self.logger.log(self.level, self.buf)



import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


# Required dependencies
required = [
    'joblib',
    'numpy',
    'tqdm',
    'humanhash3',
    'python-dateutil'
]

setup(
    name="smallab",
    version="1.4.2",
    url='https://github.com/octopuscabbage/smallab',
    packages=find_packages(),
    install_requires=required,
    license="BSD 2-Clause License",
    description="smallab (Small Lab) is an experiment framework designed " + \
                "to be easy to use with your experiment",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
)



import typing

import abc

from smallab.smallab_types import Specification


class BaseExperiment(abc.ABC):
    def set_logger_name(self, logger):
        """
        Called by ExperimentRunner to set a per-experiment logger
        :param logger: The logger which all logging in this experiment should be done to
        """
        self.logger = logger

    def get_logger_name(self):
        """
        Gets the logger that all logging in this experiment should be done to
        :return: An instance of logging.logger which corresponds to this experiment
        """
        return self.logger

    def set_logging_folder(self, folder_loc):
        """
        Set by ExperimentLogger, where the log files should go.
        :param folder_loc:
        :return:
        """
        self.logging_folder = folder_loc

    def get_logging_folder(self):
        """
        Read by ExperimentLogger, where the log files should go.
        :return:
        """
        return self.logging_folder


class Experiment(BaseExperiment):
    """
    This represents the base class for what the user needs to implement.
    The main method takes a dictionary which is the specification and returns a dictionary which is the result.
    This is run by the ExperimentRunner class

    The lifecycle of this object is
    {} is called internally by smallab
    {.set_logging_folder() -> .get_logging_folder() -> .set_logger()} -> .main()
    """

    @abc.abstractmethod
    def main(self, specification: Specification) -> typing.Dict:
        """
        The method that should be overriden when using smallab.
        Should contain the code for running your experiment.

        If an exception is raised the experiment is considered failed.

        :param specification: A dictionary passed which represents the variables of this experiment
        :return: A dictionary of results which will be serialized and saved if this experiment is succesful
        """
        pass


class CheckpointedExperiment(BaseExperiment):
    """
    CheckpointedExperiment is an Experiment which can be stopped and restarted.
    Step is called multiple times, after it returns the current experiment is serialized allowing the experiment to
    restart seamlessly since to step it will not look any different whether python has shut down in between successive calls.
    Smallab will handle serialization and deserialization of this object.
    This will fail if objects attached to the experiment are not pickleable.

    The lifecycle of this object is

    {} is called internally by smallab, []* is called multiple times
    {.set_logging_folder() -> .get_logging_folder() -> .set_logger()} -> .initialize() -> [.step()]*
    """

    @abc.abstractmethod
    def initialize(self, specification: Specification):
        pass

    @abc.abstractmethod
    def step(self) -> typing.Optional[typing.Dict]:
        """
        This method is to do work between checkpoints. 
        Intermediate results should be saved to self
        :return: None to indicate step should be called again, a dictionary of results when experiment has finished
        """
        pass



#Subclassing experiment is the simplest way to implement your experiment but is not fault tolerant
#Subclassing CheckpointedExperiment allows your experiment to fail (or the running machine to lose power) and resume at a later point
#Note: Checkpointing does have some overhead so if the experiment is very fast consider not using this
import logging
import time
import typing

import random
from numpy.random.mtrand import RandomState

from smallab.experiment import Experiment, CheckpointedExperiment
from smallab.runner import ExperimentRunner
from smallab.runner_implementations.multiprocessing_runner import MultiprocessingRunner
from smallab.smallab_types import Specification
from smallab.specification_generator import SpecificationGenerator
from examples.example_utils import delete_experiments_folder


#Similar experiment as before but now it will sleep for a bit, taking longer to complete in serial
class SimpleExperiment(CheckpointedExperiment):
    #Initialize will be called once to set up your experiment
    #Each specification will have one instance of your experiment saved with it, so save variables you need to self
    def initialize(self, specification: Specification):
        self.seed = specification['seed']
        self.num_calls = specification["num_calls"]
        logging.getLogger(self.get_logger_name()).info("Doing work!")
        self.rs = RandomState(self.seed)
        self.i = 0

    #Step will be called many times until it does not return None
    def step(self) -> typing.Optional[typing.Dict]:
        logging.getLogger(self.get_logger_name()).info("Stepping")
        if self.i < self.num_calls:
            self.r = self.rs.random()
            time.sleep(10 * self.r)
        #this experiment can have a random transient failure!
        #Since it's checkpointed, it will likely succeed after running it again
        if random.randint(0,100) > 90:
            raise Exception("Something bad happened, a moth flew into the computer!")
        if self.i >= self.num_calls:
            #Done with the experiment, return the results dictionary like normal
            return {"number": self.r}
        else:
            #This experiment isn't done, return None to signify step needs to be called again
            return


#Same specification as before
generation_specification = {"seed": [1, 2, 3, 4, 5, 6, 7, 8], "num_calls": [1, 2, 3]}
specifications = SpecificationGenerator().generate(generation_specification)

name = "checkpointed_run"
#This time we will run them all in parallel
runner = ExperimentRunner()
runner.run(name, specifications, SimpleExperiment(),specification_runner=MultiprocessingRunner())

#Some of our experiments may have failed, let's call run again to hopefully solve that
runner.run(name, specifications, SimpleExperiment(),specification_runner=MultiprocessingRunner())

#Cleanup example
delete_experiments_folder(name)


import collections
import itertools
import json
import typing

import copy


class SpecificationGenerator:
    def generate(self, generation_specification: typing.Dict) -> typing.List[collections.OrderedDict]:
        '''
        This class takes a generation specification and outputs a list of specifications based on it.
        The format is as follows
        {
            "a": [1,2,3]
            "b": [1,2]
        }

        Generates

        [{
            "a": 1,
            "b": 1
        },
        {
            "a": 2,
            "b": 1
        },
        {
            "a": 3,
            "b": 1
        },
        {
            "a": 1,
            "b": 2
        },
        {
            "a": 2,
            "b": 2
        },
        {
            "a": 3,
            "b": 2
        }
        ]

        if you want key with a list value put it in double brackets ie
        {
            "a": [[1,2,3]]
        }
        '''
        generation_specification = collections.OrderedDict(sorted(generation_specification.items()))
        iterators = []
        for key, value in generation_specification.items():
            if isinstance(value, list):
                iterators.append(list(map(lambda x: (key, x), value)))
        specifications = []
        for updates in itertools.product(*iterators):
            cur_j = copy.deepcopy(generation_specification)
            for update_key, update_value in updates:
                cur_j[update_key] = update_value
            specifications.append(cur_j)
        return specifications

    def from_json_file(self, fp: typing.AnyStr) -> typing.List[typing.Dict]:
        with open(fp) as f:
            j = json.load(f)
        if isinstance(j, list):
            out = []
            for specification in j:
                out.extend(self.generate(specification))
            return out
        else:
            return self.generate(j)


class MultiComputerGenerator(SpecificationGenerator):
    def __init__(self, computer_number, number_of_computers):
        '''
        Divides the specification across multiple computers by giving each computer the i*computer_numberth specification

        If you have 3 computers then the computer numbers would be 0,1,2 and the number_of_computers would be 3

        :param computer_number: 0 indexed number which to assign this computer (MUST NOT OVERLAP WITH ANOTHER COMPUTER)
        :param number_of_computers: The total number of computers which are being used
        '''
        assert computer_number < number_of_computers, "Computer number must be less than number of computers"

        self.computer_number = computer_number
        self.number_of_computers = number_of_computers

    def shard(self, l: typing.List) -> typing.List:
        return l[self.computer_number::self.number_of_computers]

    def generate(self, generation_specification: typing.Dict) -> typing.List[typing.Dict]:
        return self.shard(super().generate(generation_specification))



import datetime
import logging
import typing

import dill
import os
from dateutil.parser import parse

from smallab.dashboard.dashboard import eventQueue
from smallab.dashboard.dashboard_events import ProgressEvent
from smallab.dashboard.utils import put_in_event_queue
from smallab.experiment import CheckpointedExperiment
from smallab.file_locations import get_partial_save_directory
from smallab.smallab_types import Specification
from smallab.specification_hashing import specification_hash


class CheckpointedExperimentHandler():
    """
    An internal handler to handle running checkpointed experiments.
    """

    def __init__(self, rolled_backups=3):
        """

        :param rolled_backups: Controls how many backups are kept. Once this limit is reached the oldest is deleted.
        """
        self.rolled_backups = rolled_backups

    def run(self, experiment: CheckpointedExperiment, name: typing.AnyStr, specification: Specification):
        loaded_experiment = self.load_most_recent(name, specification)
        if loaded_experiment is None:
            experiment.initialize(specification)
            self.__save_checkpoint(experiment, name, specification)
        else:
            experiment = loaded_experiment
        result = experiment.step()
        self.publish_completion(specification, result)
        while result is None or isinstance(result, tuple):
            self.__save_checkpoint(experiment, name, specification)
            result = experiment.step()
            self.publish_completion(specification, result)
        return result

    def publish_completion(self, specification, result):
        if isinstance(result, tuple):
            put_in_event_queue(ProgressEvent(specification_hash(specification), result[0], result[1]))

    def load_most_recent(self, name, specification):
        specification_id = specification_hash(specification)
        location = get_partial_save_directory(name, specification)
        checkpoints = self._get_time_sorted_checkpoints(name, specification)
        if checkpoints == []:
            logging.getLogger("smallab.{id}.checkpoint".format(id=specification_id)).info("No checkpoints available")
            return
        # checkpoints = reversed(checkpoints)
        able_to_load_checkpoint = False
        checkpoints = reversed(checkpoints)
        used_checkpoint = None
        for checkpoint in checkpoints:
            try:
                with open(os.path.join(location, str(checkpoint) + ".pkl"), "rb") as f:
                    partial_experiment = dill.load(f)
                    able_to_load_checkpoint = True
                    used_checkpoint = checkpoint
                    break
            except:
                logging.getLogger("smallab.{id}.checkpoint".format(id=specification_id)).warning(
                    "Unable to load checkpoint {chp}".format(chp=checkpoint), exc_info=True)
        if not able_to_load_checkpoint:
            logging.getLogger("smallab.{id}.checkpoint".format(id=specification_id)).warning(
                "All checkpoints corrupt".format(id=specification_id))
            return
        else:
            logging.getLogger("smallab.{id}.checkpoint".format(id=specification_id)).info(
                "Successfully loaded checkpoint {chp}".format(chp=used_checkpoint))
        return partial_experiment

    def _get_time_sorted_checkpoints(self, name, specification):

        location = get_partial_save_directory(name, specification)
        try:
            checkpoints = os.listdir(location)
        except FileNotFoundError:
            return []
        checkpoints = sorted(
            list(map(parse, map(lambda x: x.strip(".pkl"), checkpoints))))
        return checkpoints

    def __save_checkpoint(self, experiment, name, specification):
        experiment_hash = specification_hash(specification)
        checkpoint_name = str(datetime.datetime.now())
        try:
            location = get_partial_save_directory(name, specification)
            os.makedirs(location, exist_ok=True)
            # TODO make sure a checkpoint with this name doesn't already exist
            with open(os.path.join(location, checkpoint_name + ".pkl"), "wb") as f:
                dill.dump(experiment, f)
            logging.getLogger("smallab.{id}.checkpoint".format(id=experiment_hash)).info(
                "Succesfully checkpointed {chp}".format(chp=checkpoint_name))
            checkpoints = os.listdir(location)
            if len(checkpoints) > self.rolled_backups:
                checkpoints = self._get_time_sorted_checkpoints(name, specification)
                os.remove(os.path.join(location, str(checkpoints[0]) + ".pkl"))
        except:
            logging.getLogger("smallab.{id}.checkpoint".format(id=experiment_hash)).warning(
                "Unsuccesful checkpoint {chp}".format(chp=checkpoint_name),
                exc_info=True)



import json
import typing

import os

from smallab.specification_hashing import specification_hash
from smallab.smallab_types import Specification

experiment_folder = "experiment_runs"


def get_save_directory(name: typing.AnyStr) -> typing.AnyStr:
    """
    Get the directory that the runs will be saved under
    :param name: The name of the current batch
    :return: The folder to save the runs under
    """
    return os.path.join(experiment_folder, name)


def get_experiment_save_directory(name):
    return os.path.join(get_save_directory(name), "experiments")


def get_pkl_file_location(name: typing.AnyStr,
                          specification: Specification) -> typing.AnyStr:
    """
    Get the filename to save the file under
    :param name: The name of the current batch
    :param specification: The specification of the current run
    :return: The filename to save this run under
    """
    return os.path.join(get_save_file_directory(name, specification),"run.pkl")


def get_json_file_location(name, specification):
    return os.path.join(get_save_file_directory(name, specification),"run.json")


def get_specification_file_location(name: typing.AnyStr, specification: typing.Dict) -> typing.AnyStr:
    """
    Get the specification file location
    :param name: The name of the current batch
    :param specification: The specification of the current run
    :return: The location where the specification.json should be saved
    """
    return os.path.join(get_save_file_directory(name, specification), "specification.json")


def get_save_file_directory(name: typing.AnyStr, specification: Specification) -> typing.AnyStr:
    """
    Get the folder to save the .pkl file and specification.json file under
    :param name: The name of the current batch
    :param specification: The specification of the current run
    :return: The location where specification.json should be saved
    """
    return os.path.join(get_experiment_save_directory(name), specification_hash(specification))


def get_partial_save_directory(name, specification):
    return os.path.join(get_save_directory(name), "checkpoints", specification_hash(specification))


def get_log_file(experiment, specification_id):
    return os.path.join(experiment.get_logging_folder(),specification_id + ".log")



import typing

import abc

from smallab.utilities.hooks import format_exception


class CallbackManager(abc.ABC):
    def set_experiment_name(self, name):
        self.name = name

    def on_specification_complete(self, specification: typing.Dict, result: typing.Dict) -> typing.NoReturn:
        """
        Sets a method to be called on the completion of each experiment

        :return: No return
        """
        pass

    def on_specification_failure(self, exception: Exception, specification: typing.Dict) -> typing.NoReturn:
        """
        Sets a method to be called on the failure of each experiment
        :return: No Return
        """
        pass

    def on_batch_complete(self, specifications: typing.List[typing.Dict]) -> typing.NoReturn:
        """
        Sets a method to be called on the completion of all specifications provided to run
        :return: No return
        """
        pass

    def on_batch_failure(self, errors: typing.List[Exception],
                         specifications: typing.List[typing.Dict]) -> typing.NoReturn:
        """
        Sets a method to be called when 1 or more specifications failed to complete (Threw an exception
        :return: No return
        """
        pass


class PrintCallback(CallbackManager):
    def on_specification_complete(self, specification, return_value) -> typing.NoReturn:
        print("!!! Complete !!! ")
        print(specification)
        print(return_value)

    def on_specification_failure(self, exception, specification) -> typing.NoReturn:
        print("!!! Failure !!!")
        print(specification)
        print(format_exception(exception))
